@@ -17,36 +17,52 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
+from functools import partial
 import os
 import re
-import sys
-from functools import partial
-from PyQt4 import QtCore, QtGui
-from picard import config, log
-from picard.album import Album, NatAlbum
-from picard.cluster import Cluster, ClusterList, UnmatchedFiles
+
+from PyQt5 import (
+    QtCore,
+    QtGui,
+    QtWidgets,
+)
+
+from picard import (
+    config,
+    log,
+)
+from picard.album import (
+    Album,
+    NatAlbum,
+)
+from picard.cluster import (
+    Cluster,
+    ClusterList,
+    UnclusteredFiles,
+)
 from picard.file import File
-from picard.track import Track, NonAlbumTrack
-from picard.util import encode_filename, icontheme
 from picard.plugin import ExtensionPoint
-from picard.ui.ratingwidget import RatingWidget
+from picard.track import (
+    NonAlbumTrack,
+    Track,
+)
+from picard.util import (
+    encode_filename,
+    icontheme,
+    restore_method,
+)
+
 from picard.ui.collectionmenu import CollectionMenu
-
-if sys.platform == 'darwin':
-    try:
-        from Foundation import NSURL
-        NSURL_IMPORTED = True
-    except ImportError:
-        NSURL_IMPORTED = False
-        log.warning("Unable to import NSURL, file drag'n'drop might not work correctly")
+from picard.ui.ratingwidget import RatingWidget
+from picard.ui.scriptsmenu import ScriptsMenu
 
 
-class BaseAction(QtGui.QAction):
+class BaseAction(QtWidgets.QAction):
     NAME = "Unknown"
     MENU = []
 
     def __init__(self):
-        QtGui.QAction.__init__(self, self.NAME, None)
+        super().__init__(self.NAME, None)
         self.triggered.connect(self.__callback)
 
     def __callback(self):
@@ -93,7 +109,7 @@ def get_match_color(similarity, basecolor):
         c2[2] + (c1[2] - c2[2]) * similarity)
 
 
-class MainPanel(QtGui.QSplitter):
+class MainPanel(QtWidgets.QSplitter):
 
     options = [
         config.Option("persist", "splitter_state", QtCore.QByteArray()),
@@ -106,7 +122,7 @@ class MainPanel(QtGui.QSplitter):
     ]
 
     def __init__(self, window, parent=None):
-        QtGui.QSplitter.__init__(self, parent)
+        super().__init__(parent)
         self.window = window
         self.create_icons()
         self.views = [FileTreeView(window, self), AlbumTreeView(window, self)]
@@ -128,7 +144,7 @@ class MainPanel(QtGui.QSplitter):
         }
         FileItem.file_colors = {
             File.NORMAL: TreeItem.text_color,
-            File.CHANGED: config.setting["color_modified"],
+            File.CHANGED: TreeItem.text_color,
             File.PENDING: config.setting["color_pending"],
             File.ERROR: config.setting["color_error"],
         }
@@ -138,12 +154,13 @@ class MainPanel(QtGui.QSplitter):
         for view in self.views:
             view.save_state()
 
+    @restore_method
     def restore_state(self):
         self.restoreState(config.persist["splitter_state"])
 
     def create_icons(self):
-        if hasattr(QtGui.QStyle, 'SP_DirIcon'):
-            ClusterItem.icon_dir = self.style().standardIcon(QtGui.QStyle.SP_DirIcon)
+        if hasattr(QtWidgets.QStyle, 'SP_DirIcon'):
+            ClusterItem.icon_dir = self.style().standardIcon(QtWidgets.QStyle.SP_DirIcon)
         else:
             ClusterItem.icon_dir = icontheme.lookup('folder', icontheme.ICON_SIZE_MENU)
         AlbumItem.icon_cd = icontheme.lookup('media-optical', icontheme.ICON_SIZE_MENU)
@@ -155,6 +172,7 @@ class MainPanel(QtGui.QSplitter):
         TrackItem.icon_audio = QtGui.QIcon(":/images/track-audio.png")
         TrackItem.icon_video = QtGui.QIcon(":/images/track-video.png")
         TrackItem.icon_data = QtGui.QIcon(":/images/track-data.png")
+        TrackItem.icon_error = icontheme.lookup('dialog-error', icontheme.ICON_SIZE_MENU)
         FileItem.icon_file = QtGui.QIcon(":/images/file.png")
         FileItem.icon_file_pending = QtGui.QIcon(":/images/file-pending.png")
         FileItem.icon_error = icontheme.lookup('dialog-error', icontheme.ICON_SIZE_MENU)
@@ -220,17 +238,16 @@ class MainPanel(QtGui.QSplitter):
             self.update_current_view()
 
 
-class BaseTreeView(QtGui.QTreeWidget):
+class BaseTreeView(QtWidgets.QTreeWidget):
 
     options = [
-        config.Option("setting", "color_modified", QtGui.QColor(QtGui.QPalette.WindowText)),
         config.Option("setting", "color_saved", QtGui.QColor(0, 128, 0)),
         config.Option("setting", "color_error", QtGui.QColor(200, 0, 0)),
         config.Option("setting", "color_pending", QtGui.QColor(128, 128, 128)),
     ]
 
     def __init__(self, window, parent=None):
-        QtGui.QTreeWidget.__init__(self, parent)
+        super().__init__(parent)
         self.window = window
         self.panel = parent
 
@@ -241,7 +258,7 @@ class BaseTreeView(QtGui.QTreeWidget):
         self.setAcceptDrops(True)
         self.setDragEnabled(True)
         self.setDropIndicatorShown(True)
-        self.setSelectionMode(QtGui.QAbstractItemView.ExtendedSelection)
+        self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
 
         # enable sorting, but don't actually use it by default
         # XXX it would be nice to be able to go to the 'no sort' mode, but the
@@ -249,13 +266,13 @@ class BaseTreeView(QtGui.QTreeWidget):
         self.header().setSortIndicator(-1, QtCore.Qt.AscendingOrder)
         self.setSortingEnabled(True)
 
-        self.expand_all_action = QtGui.QAction(_("&Expand all"), self)
+        self.expand_all_action = QtWidgets.QAction(_("&Expand all"), self)
         self.expand_all_action.triggered.connect(self.expandAll)
-        self.collapse_all_action = QtGui.QAction(_("&Collapse all"), self)
+        self.collapse_all_action = QtWidgets.QAction(_("&Collapse all"), self)
         self.collapse_all_action.triggered.connect(self.collapseAll)
-        self.select_all_action = QtGui.QAction(_("Select &all"), self)
+        self.select_all_action = QtWidgets.QAction(_("Select &all"), self)
         self.select_all_action.triggered.connect(self.selectAll)
-        self.select_all_action.setShortcut(QtGui.QKeySequence(_(u"Ctrl+A")))
+        self.select_all_action.setShortcut(QtGui.QKeySequence(_("Ctrl+A")))
         self.doubleClicked.connect(self.activate_item)
 
     def contextMenuEvent(self, event):
@@ -265,7 +282,7 @@ class BaseTreeView(QtGui.QTreeWidget):
         obj = item.obj
         plugin_actions = None
         can_view_info = self.window.view_info_action.isEnabled()
-        menu = QtGui.QMenu(self)
+        menu = QtWidgets.QMenu(self)
 
         if isinstance(obj, Track):
             if can_view_info:
@@ -287,7 +304,7 @@ class BaseTreeView(QtGui.QTreeWidget):
             menu.addSeparator()
             menu.addAction(self.window.autotag_action)
             menu.addAction(self.window.analyze_action)
-            if isinstance(obj, UnmatchedFiles):
+            if isinstance(obj, UnclusteredFiles):
                 menu.addAction(self.window.cluster_action)
             else:
                 menu.addAction(self.window.album_search_action)
@@ -321,7 +338,7 @@ class BaseTreeView(QtGui.QTreeWidget):
         bottom_separator = False
 
         if isinstance(obj, Album) and not isinstance(obj, NatAlbum) and obj.loaded:
-            releases_menu = QtGui.QMenu(_("&Other versions"), menu)
+            releases_menu = QtWidgets.QMenu(_("&Other versions"), menu)
             menu.addSeparator()
             menu.addMenu(releases_menu)
             loading = releases_menu.addAction(_('Loading...'))
@@ -347,8 +364,8 @@ class BaseTreeView(QtGui.QTreeWidget):
                     for version in versions:
                         priority = {
                             "trackmatch": "0" if version['totaltracks'] == albumtracks else "?",
-                            "countrymatch": "0" if len(preferred_countries) == 0 or preferred_countries & set(version['countries']) else "?",
-                            "formatmatch": "0" if len(preferred_formats) == 0 or preferred_formats & set(version['formats']) else "?",
+                            "countrymatch": "0" if len(preferred_countries) == 0 or preferred_countries & set(version['countries'] or '') else "?",
+                            "formatmatch": "0" if len(preferred_formats) == 0 or preferred_formats & set(version['formats'] or '') else "?",
                         }
                         priorities[version['id']] = "".join(priority[k] for k in matches)
                     versions.sort(key=lambda version: priorities[version['id']] + version['name'])
@@ -378,7 +395,7 @@ class BaseTreeView(QtGui.QTreeWidget):
         if config.setting["enable_ratings"] and \
            len(self.window.selected_objects) == 1 and isinstance(obj, Track):
             menu.addSeparator()
-            action = QtGui.QWidgetAction(menu)
+            action = QtWidgets.QWidgetAction(menu)
             action.setDefaultWidget(RatingWidget(menu, obj))
             menu.addAction(action)
             menu.addSeparator()
@@ -391,16 +408,20 @@ class BaseTreeView(QtGui.QTreeWidget):
                 menu.addSeparator()
             menu.addMenu(CollectionMenu(selected_albums, _("Collections"), menu))
 
-        if plugin_actions:
-            plugin_menu = QtGui.QMenu(_("P&lugins"), menu)
-            plugin_menu.setIcon(self.panel.icon_plugins)
+        scripts = config.setting["list_of_scripts"]
+
+        if plugin_actions or scripts:
             menu.addSeparator()
+
+        if plugin_actions:
+            plugin_menu = QtWidgets.QMenu(_("P&lugins"), menu)
+            plugin_menu.setIcon(self.panel.icon_plugins)
             menu.addMenu(plugin_menu)
 
             plugin_menus = {}
             for action in plugin_actions:
                 action_menu = plugin_menu
-                for index in xrange(1, len(action.MENU) + 1):
+                for index in range(1, len(action.MENU) + 1):
                     key = tuple(action.MENU[:index])
                     if key in plugin_menus:
                         action_menu = plugin_menus[key]
@@ -408,6 +429,11 @@ class BaseTreeView(QtGui.QTreeWidget):
                         action_menu = plugin_menus[key] = action_menu.addMenu(key[-1])
                 action_menu.addAction(action)
 
+        if scripts:
+            scripts_menu = ScriptsMenu(scripts, _("&Run scripts"), menu)
+            scripts_menu.setIcon(self.panel.icon_plugins)
+            menu.addMenu(scripts_menu)
+
         if isinstance(obj, Cluster) or isinstance(obj, ClusterList) or isinstance(obj, Album):
             menu.addSeparator()
             menu.addAction(self.expand_all_action)
@@ -417,6 +443,7 @@ class BaseTreeView(QtGui.QTreeWidget):
         menu.exec_(event.globalPos())
         event.accept()
 
+    @restore_method
     def restore_state(self):
         sizes = config.persist[self.view_sizes.name]
         header = self.header()
@@ -452,7 +479,12 @@ class BaseTreeView(QtGui.QTreeWidget):
         if items:
             drag = QtGui.QDrag(self)
             drag.setMimeData(self.mimeData(items))
-            drag.start(supportedActions)
+            # Render the first selected element as drag representation
+            rectangle = self.visualItemRect(items[0])
+            pixmap = QtGui.QPixmap(rectangle.width(), rectangle.height())
+            self.viewport().render(pixmap, QtCore.QPoint(), QtGui.QRegion(rectangle))
+            drag.setPixmap(pixmap)
+            drag.exec_(QtCore.Qt.MoveAction)
 
     def mimeData(self, items):
         """Return MIME data for specified items."""
@@ -462,11 +494,11 @@ class BaseTreeView(QtGui.QTreeWidget):
         for item in items:
             obj = item.obj
             if isinstance(obj, Album):
-                album_ids.append(str(obj.id))
+                album_ids.append(obj.id)
             elif obj.iterfiles:
                 files.extend([url(f.filename) for f in obj.iterfiles()])
         mimeData = QtCore.QMimeData()
-        mimeData.setData("application/picard.album-list", "\n".join(album_ids))
+        mimeData.setData("application/picard.album-list", "\n".join(album_ids).encode())
         if files:
             mimeData.setUrls(files)
         return mimeData
@@ -478,18 +510,7 @@ class BaseTreeView(QtGui.QTreeWidget):
         for url in urls:
             log.debug("Dropped the URL: %r", url.toString(QtCore.QUrl.RemoveUserInfo))
             if url.scheme() == "file" or not url.scheme():
-                if sys.platform == 'darwin' and unicode(url.path()).startswith('/.file/id='):
-                    # Workaround for https://bugreports.qt.io/browse/QTBUG-40449
-                    # OSX Urls follow the NSURL scheme and need to be converted
-                    if NSURL_IMPORTED:
-                        filename = os.path.normpath(os.path.realpath(unicode(NSURL.URLWithString_(str(url.toString())).filePathURL().path()).rstrip("\0")))
-                        log.debug('OSX NSURL path detected. Dropped File is: %r', filename)
-                    else:
-                        log.error("Unable to get appropriate file path for %r", url.toString(QtCore.QUrl.RemoveUserInfo))
-                        continue
-                else:
-                    # Dropping a file from iTunes gives a filename with a NULL terminator
-                    filename = os.path.normpath(os.path.realpath(unicode(url.toLocalFile()).rstrip("\0")))
+                filename = os.path.normpath(os.path.realpath(url.toLocalFile().rstrip("\0")))
                 file = BaseTreeView.tagger.files.get(filename)
                 if file:
                     files.append(file)
@@ -498,7 +519,7 @@ class BaseTreeView(QtGui.QTreeWidget):
                 else:
                     new_files.append(filename)
             elif url.scheme() in ("http", "https"):
-                path = unicode(url.path())
+                path = url.path()
                 match = re.search(r"/(release|recording)/([0-9a-z\-]{36})", path)
                 if match:
                     entity = match.group(1)
@@ -513,7 +534,7 @@ class BaseTreeView(QtGui.QTreeWidget):
             BaseTreeView.tagger.add_files(new_files, target=target)
 
     def dropEvent(self, event):
-        return QtGui.QTreeView.dropEvent(self, event)
+        return QtWidgets.QTreeView.dropEvent(self, event)
 
     def dropMimeData(self, parent, index, data, action):
         target = None
@@ -535,8 +556,8 @@ class BaseTreeView(QtGui.QTreeWidget):
         albums = data.data("application/picard.album-list")
         if albums:
             if isinstance(self, FileTreeView) and target is None:
-                target = self.tagger.unmatched_files
-            albums = [self.tagger.load_album(id) for id in str(albums).split("\n")]
+                target = self.tagger.unclustered_files
+            albums = [self.tagger.load_album(id) for id in bytes(albums).decode().split("\n")]
             self.tagger.move_files(self.tagger.get_files_from_objects(albums), target)
             handled = True
         return handled
@@ -559,11 +580,11 @@ class BaseTreeView(QtGui.QTreeWidget):
             cluster_item.add_files(cluster.files)
 
     def moveCursor(self, action, modifiers):
-        if action in (QtGui.QAbstractItemView.MoveUp, QtGui.QAbstractItemView.MoveDown):
+        if action in (QtWidgets.QAbstractItemView.MoveUp, QtWidgets.QAbstractItemView.MoveDown):
             item = self.currentItem()
             if item and not item.isSelected():
                 self.setCurrentItem(item)
-        return QtGui.QTreeWidget.moveCursor(self, action, modifiers)
+        return QtWidgets.QTreeWidget.moveCursor(self, action, modifiers)
 
 
 class FileTreeView(BaseTreeView):
@@ -571,15 +592,15 @@ class FileTreeView(BaseTreeView):
     view_sizes = config.TextOption("persist", "file_view_sizes", "250 40 100")
 
     def __init__(self, window, parent=None):
-        BaseTreeView.__init__(self, window, parent)
+        super().__init__(window, parent)
         self.setAccessibleName(_("file view"))
         self.setAccessibleDescription(_("Contains unmatched files and clusters"))
-        self.unmatched_files = ClusterItem(self.tagger.unmatched_files, False, self)
+        self.unmatched_files = ClusterItem(self.tagger.unclustered_files, False, self)
         self.unmatched_files.update()
-        self.setItemExpanded(self.unmatched_files, True)
+        self.unmatched_files.setExpanded(True)
         self.clusters = ClusterItem(self.tagger.clusters, False, self)
         self.set_clusters_text()
-        self.setItemExpanded(self.clusters, True)
+        self.clusters.setExpanded(True)
         self.tagger.cluster_added.connect(self.add_file_cluster)
         self.tagger.cluster_removed.connect(self.remove_file_cluster)
 
@@ -593,7 +614,7 @@ class FileTreeView(BaseTreeView):
         self.set_clusters_text()
 
     def set_clusters_text(self):
-        self.clusters.setText(0, '%s (%d)' % (_(u"Clusters"), len(self.tagger.clusters)))
+        self.clusters.setText(0, '%s (%d)' % (_("Clusters"), len(self.tagger.clusters)))
 
 
 class AlbumTreeView(BaseTreeView):
@@ -601,7 +622,7 @@ class AlbumTreeView(BaseTreeView):
     view_sizes = config.TextOption("persist", "album_view_sizes", "250 40 100")
 
     def __init__(self, window, parent=None):
-        BaseTreeView.__init__(self, window, parent)
+        super().__init__(window, parent)
         self.setAccessibleName(_("album view"))
         self.setAccessibleDescription(_("Contains albums and matched files"))
         self.tagger.album_added.connect(self.add_album)
@@ -622,20 +643,19 @@ class AlbumTreeView(BaseTreeView):
         self.takeTopLevelItem(self.indexOfTopLevelItem(album.item))
 
 
-class TreeItem(QtGui.QTreeWidgetItem):
-
-    __lt__ = lambda self, other: False
+class TreeItem(QtWidgets.QTreeWidgetItem):
 
     def __init__(self, obj, sortable, *args):
-        QtGui.QTreeWidgetItem.__init__(self, *args)
+        super().__init__(*args)
         self.obj = obj
         if obj is not None:
             obj.item = self
-        if sortable:
-            self.__lt__ = self._lt
+        self.sortable = sortable
         self.setTextAlignment(1, QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
 
-    def _lt(self, other):
+    def __lt__(self, other):
+        if not self.sortable:
+            return False
         column = self.treeWidget().sortColumn()
         if column == 1:
             return (self.obj.metadata.length or 0) < (other.obj.metadata.length or 0)
@@ -645,7 +665,7 @@ class TreeItem(QtGui.QTreeWidgetItem):
 class ClusterItem(TreeItem):
 
     def __init__(self, *args):
-        TreeItem.__init__(self, *args)
+        super().__init__(*args)
         self.setIcon(0, ClusterItem.icon_dir)
 
     def update(self):
@@ -688,19 +708,20 @@ class AlbumItem(TreeItem):
             oldnum = self.childCount() - 1
             newnum = len(album.tracks)
             if oldnum > newnum:  # remove old items
-                for i in xrange(oldnum - newnum):
-                    self.takeChild(newnum - 1)
+                for i in range(oldnum - newnum):
+                    self.takeChild(newnum)
                 oldnum = newnum
             # update existing items
-            for i in xrange(oldnum):
+            for i in range(oldnum):
                 item = self.child(i)
+                item.setSelected(False)
                 track = album.tracks[i]
                 item.obj = track
                 track.item = item
                 item.update(update_album=False)
             if newnum > oldnum:  # add new items
                 items = []
-                for i in xrange(newnum - 1, oldnum - 1, -1):  # insertChildren is backwards
+                for i in range(newnum - 1, oldnum - 1, -1):  # insertChildren is backwards
                     item = TrackItem(album.tracks[i], False)
                     item.setHidden(False)  # Workaround to make sure the parent state gets updated
                     items.append(item)
@@ -742,6 +763,7 @@ class TrackItem(TreeItem):
             icon = FileItem.decide_file_icon(file)
             self.setToolTip(0, _(FileItem.decide_file_icon_info(file)))
             self.takeChildren()
+            self.setExpanded(False)
         else:
             self.setToolTip(0, "")
             if track.ignored_for_completeness():
@@ -759,10 +781,10 @@ class TrackItem(TreeItem):
                 oldnum = self.childCount()
                 newnum = track.num_linked_files
                 if oldnum > newnum:  # remove old items
-                    for i in xrange(oldnum - newnum):
+                    for i in range(oldnum - newnum):
                         self.takeChild(newnum - 1).obj.item = None
                     oldnum = newnum
-                for i in xrange(oldnum):  # update existing items
+                for i in range(oldnum):  # update existing items
                     item = self.child(i)
                     file = track.linked_files[i]
                     item.obj = file
@@ -770,13 +792,17 @@ class TrackItem(TreeItem):
                     item.update(update_track=False)
                 if newnum > oldnum:  # add new items
                     items = []
-                    for i in xrange(newnum - 1, oldnum - 1, -1):
+                    for i in range(newnum - 1, oldnum - 1, -1):
                         item = FileItem(track.linked_files[i], False)
                         item.update(update_track=False)
                         items.append(item)
                     self.addChildren(items)
             self.setExpanded(True)
-        self.setIcon(0, icon)
+        if track.error:
+            self.setIcon(0, TrackItem.icon_error)
+            self.setToolTip(0, track.error)
+        else:
+            self.setIcon(0, icon)
         for i, column in enumerate(MainPanel.columns):
             self.setText(i, track.column(column[1]))
             self.setForeground(i, color)
